# case.rb - pattern matching for Ruby
#
# Copyright 2008  MenTaLguY <mental@rydia.net>
#
# This library is made available under the same terms as Ruby.
#

module Case

module Impl #:nodoc: all
  module Struct
    def ===(obj)
      return false unless self.class === obj
      zip(obj) { |e,o| return false unless e === o }
      true
    end
  end

  module Array
    def ===(obj)
      return false unless ::Array === obj
      zip(obj) { |e,o| return false unless e === o }
      true
    end
  end

  class Operator
    class << self
      alias_method :[], :new
    end

    def initialize(*patterns)
      @patterns = patterns
    end

    def ===(obj)
      raise NotImplementedError, "#{self.class}#=== not implemented"
    end
  end
end

class Not
  class << self
    alias_method :[], :new
  end

  def initialize(pattern)
    @pattern = pattern
  end

  def ===(obj)
    !( @pattern === obj )
  end
end

class Any < Impl::Operator
  def self.===(obj) ; true ; end

  def ===(obj)
    @patterns.any? { |p| p === obj }
  end
end

class All < Impl::Operator
  def ===(obj)
    @patterns.all? { |p| p === obj }
  end
end

class Struct < ::Struct
  include Impl::Struct
end

class Array < ::Array
  include Impl::Array
end

class Guard
  def initialize(&predicate)
    @predicate = predicate
  end

  def ===(obj)
    @predicate.call obj
  end
end

def self.guard(&predicate)
  Guard.new &predicate
end

Cmp = Object.new
for op in %w(< > <= >=)
  eval <<-EOS
    def Cmp.#{op}(other)
      Guard.new { |value| value #{op} other }
    end
  EOS
end

# Shorthand for Case::Array.[]
def self.[](*args)
  Array[*args]
end

end
